// $Id: PnfsId.java 16563 2011-11-05 10:43:03Z behrmann $

package diskCacheV111.util;

import java.io.Serializable;
import java.util.Arrays;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Immutable representation of a pnfsId
 */
public class PnfsId implements Serializable, Comparable<PnfsId> {

    private static final String SIMPLE_PNFS_ID_REGEX = "\\p{XDigit}{1,24}";
    // FIXME STAR_PNFS_ID_REGEX is too permissive; overall length should be 24 or less.
    private static final String STAR_PNFS_ID_REGEX = "\\p{XDigit}{1,22}\\*\\p{XDigit}{1,22}";
    private static final String PNFS_ID_REGEX = "("+SIMPLE_PNFS_ID_REGEX+"|"+STAR_PNFS_ID_REGEX+")";
    private static final String PNFS_STRING_REGEX = "("+PNFS_ID_REGEX+"(\\..*)?)";
    private static final String CHIMERA_ID_REGEX = "\\p{XDigit}{36}";
    private static final String VALID_ID_REGEX = "^("+PNFS_STRING_REGEX+"|"+CHIMERA_ID_REGEX+")$";
    private static final Pattern VALID_ID_PATTERN = Pattern.compile( VALID_ID_REGEX);

    private final static int OLD_ID_SIZE = 12; // original pnfs
    private final static int NEW_ID_SIZE = 18; // chimera

    private final byte[] _a;
    private final String _domain;

    private static final long serialVersionUID = -112220393521303857L;

    public static boolean isValid( String id) {
        Matcher m = VALID_ID_PATTERN.matcher( id);
        return m.matches();
    }

    public PnfsId(String id, String domain) {
        this(_stringToBytes(id), domain);
    }

    public PnfsId(String s) {
        this(stringToId(s), stringToDomain(s));
    }

    public PnfsId(byte[] id) {
        this(id, null);
    }

    public PnfsId(byte[] id, String domain) {
        int length = id.length;
        if (length != OLD_ID_SIZE && length != NEW_ID_SIZE) {
            throw new IllegalArgumentException("Illegal pnfsid string length");
        }
        _a = new byte[length];
        System.arraycopy(id, 0, _a, 0, length);

        _domain = (domain != null) ? domain.intern() : null;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }

        if (!(o instanceof PnfsId)) {
            return false;
        }

        PnfsId other = (PnfsId) o;
        return Arrays.equals(_a, other._a)
            && (_domain == other._domain
                || (_domain != null && _domain.equals(other._domain)));
    }

    @Override
    public int hashCode() {
        return Arrays.hashCode(_a) ^ ((_domain == null) ? 0 : _domain.hashCode());
    }

    public int compareTo(PnfsId pnfsId) {
        if( pnfsId == this ) return 0;

        int i = 0;
        for (i = 0; (i < _a.length) && (_a[i] == pnfsId._a[i]); i++)
            ;
        if (i == _a.length)
            return 0;
        int t = _a[i] < 0 ? 256 + _a[i] : _a[i];
        int o = pnfsId._a[i] < 0 ? 256 + pnfsId._a[i] : pnfsId._a[i];

        return t < o ? -1 : 1;
    }

    public int getDatabaseId() {
        return (((_a[0]) & 0xFF) << 8) | ((_a[1]) & 0xFF);
    }

    public String getDomain() {
        return _domain;
    }

    public String getId() {
        return bytesToHexString(_a);
    }

    @Override
    public String toString() {
        return getId() + ((_domain != null) ? "." + _domain : "");
    }

    public String toIdString() {
        return getId();
    }

    public static String toCompleteId(String shortId) {
        return bytesToHexString(_stringToBytes(shortId));
    }

    public byte[] getBytes() {
        byte[] x = new byte[_a.length];
        System.arraycopy(_a, 0, x, 0, _a.length);
        return x;
    }

    public String toShortString() {
        StringBuilder sb = new StringBuilder();
        int i = 0;
        for (i = 0; i < 2; i++)
            sb.append(byteToHexString(_a[i]));
        for (; (i < _a.length) && (_a[i] == 0); i++)
            ;
        for (; i < _a.length; i++)
            sb.append(byteToHexString(_a[i]));
        return sb.toString();
    }

    private static final byte[] stringToId(String s) {
        int i = s.indexOf('.');
        if (i < 0) {
            return  _stringToBytes(s);
        } else {
            return _stringToBytes(s.substring(0, i));
        }
    }

    private static final String stringToDomain(String s) {
        int i = s.indexOf('.');
        if (i < 0 || i == s.length() - 1) {
            return null;
        } else {
            return s.substring(i + 1);
        }
    }

    private static String byteToHexString(byte b) {
        String s = Integer.toHexString((b < 0) ? (256 + b) : (int) b)
            .toUpperCase();
        if (s.length() == 1)
            return "0" + s;
        else
            return s;
    }

    /**
     * Translation table used by bytesToHexString.
     */
    private static final char valueToHex[] = {
        '0', '1', '2', '3', '4', '5',
        '6', '7', '8', '9', 'A', 'B',
        'C', 'D', 'E', 'F' };

    /**
     * Converts a byte array into the string representation in base 16.
     */
    private static String bytesToHexString(byte[] b) {
        char result[] = new char[2 * b.length];
        for (int i = 0; i < b.length; i++) {
            int value = (b[i] + 0x100) & 0xFF;
            result[2 * i] = valueToHex[value >> 4];
            result[2 * i + 1] = valueToHex[value & 0x0F];
        }
        return new String(result);
    }

    private static byte[] _stringToBytes(String idString) {

        int len = idString.length();
        int idSize = 0;

        switch (len) {
        case OLD_ID_SIZE * 2: // old pnfsid
            idSize = OLD_ID_SIZE;
            break;
        case NEW_ID_SIZE * 2: // himera
            idSize = NEW_ID_SIZE;
            break;
        default:
            // all id's shorter than 24 characters will be extended to 24
            if ((len > OLD_ID_SIZE * 2) || (len == 0)) {
                throw new IllegalArgumentException(
                                                   "Illegal pnfsid string length");
            }
            idSize = OLD_ID_SIZE;
        }

        byte[] a = new byte[idSize];
        int p = idString.indexOf('*');
        if (p > -1) {
            if ((p == 0) || (p == (idString.length() - 1))) {
                throw new IllegalArgumentException("Illegal use of *");
            }
            int diff = 2 * OLD_ID_SIZE - idString.length() + 1;
            StringBuilder sb = new StringBuilder();
            sb.append(idString.substring(0, p));
            for (int i = 0; i < diff; i++)
                sb.append("0");
            sb.append(idString.substring(p + 1));
            idString = sb.toString();
        }
        if (idString.length() > (2 * a.length)) {
            throw new IllegalArgumentException("Illegal pnfsid string length");
        } else if (idString.length() < (2 * a.length)) {
            StringBuilder sb = new StringBuilder();
            int m = 2 * OLD_ID_SIZE - idString.length();
            for (int i = 0; i < m; i++)
                sb.append("0");
            sb.append(idString);
            idString = sb.toString();
        }
        for (int i = 0; i < idSize; i++) {
            int l = Integer
                .parseInt(idString.substring(2 * i, 2 * (i + 1)), 16);
            a[i] = (byte) ((l < 128) ? l : (l - 256));
        }
        return a;
    }

    /**
     * Converts string representation of pnfsid into its internal binary form
     *
     * @return pnfsid as byte array
     */
    public byte[] toBinPnfsId() {
        switch (_a.length) {
        case OLD_ID_SIZE: // old pnfsid
            return PnfsIdUtil.toBinPnfsId(getId());
        case NEW_ID_SIZE: // himera
            return getBytes();
        default:
            return null;
        }
    }

    public static void main(String[] args) {
        if (args.length < 1) {
            System.out.println("USAGE : ... <pnfsId>");
            System.exit(4);
        }
        try {
            PnfsId id = new PnfsId(args[0]);
            System.out.println("id.toString()      " + id);
            System.out.println("id.getId()         " + id.getId());
            System.out.println("db.getDatabaseId() " + id.getDatabaseId());
            System.out.println("db.getDomain()     " + id.getDomain());
            System.out.println("id.getBytes()      " + java.util.Arrays.toString(id.getBytes()));
            System.out.println("id.toBinPnfsId()   " + java.util.Arrays.toString(id.toBinPnfsId()));
            System.out.println("toStringPnfsId(id.toBinPnfsId()) " + PnfsIdUtil.toStringPnfsId(id.toBinPnfsId()));
            System.exit(0);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(4);
        }
    }
}
